///////////////////////////////////////////////////////////////////////
//  Cookie.cpp: Implementation of Cookie Class
//  Encapsulates RFC 2109 and Netscape Old Specification Cookies
//
//  Created by Babu George Padamadan
//  Created on Oct 2000
//  Contact : bgeorge@wol.com or babu_george@hotmail.com
//
///////////////////////////////////////////////////////////////////////

#pragma warning (disable:4786)

#include <windows.h>
#include <wininet.h>
#include <cassert>
#include <ctime>
#include "Cookie.h"
#include "CookieUtil.h"

using namespace std;
using util::toLowerStr;

// Attribute Names
const char* OLDSPEC_ATTRIBUTES[] = {   "domain",    "path",    "secure", 
                                       "expires",   NULL 
                                   };
const char* RFC2109_ATTRIBUTES[] = {   "domain",    "path",    "secure", 
                                       "max-age",  "comment",  "version",
                                       NULL 
                                   };

bool Cookie::operator == (const Cookie& objCookie)
{
    COOKIE_MAP::const_iterator it;
    for (it = objCookie.m_mapNameValue.begin();
            it != objCookie.m_mapNameValue.end(); it++)
    {
        // Check for "Domain", "Path" and "Name/Values"
        if ((*it).first == "domain" || (*it).first == "path")
        {
            if (toLowerStr(getValue((*it).first)) != toLowerStr((*it).second))
                return false;
        }
        else if (!isAttributeName((*it).first))
        {
            if (toLowerStr(getValue((*it).first)) != toLowerStr((*it).second))
                return false;
        }
    }
    return true; 
}

bool Cookie::operator < (const Cookie& objCookie)
{ 
    // For sorting - sort with Domain and Path values
    // with the most prominent path first 
    // ie., "/" will appear after "/start"
    COOKIE_MAP::const_iterator it;
    for (it = objCookie.m_mapNameValue.begin();
            it != objCookie.m_mapNameValue.end(); it++)
    {
        if ((*it).first == "domain" || (*it).first == "path")
        {
            if (toLowerStr(getValue((*it).first)) <= toLowerStr((*it).second))
                return false;
        }
    }
    return true; 
}

bool Cookie::isEqual(const Cookie& objCookie) const
{
    // If both the cookies have the same path value,
    // same domain value and the same cookie name, they
    // are equal
    if (isRFC2109Based() != objCookie.isRFC2109Based())
        return false; // Of Different specs

    // Domain
    if (toLowerStr(objCookie.getDomain()) != toLowerStr(getDomain()))
        return false;
    // Path
    if (toLowerStr(objCookie.getPath()) != toLowerStr(getPath()))
        return false;

    COOKIE_MAP::const_iterator it;
    for (it = objCookie.m_mapNameValue.begin();
            it != objCookie.m_mapNameValue.end(); it++)
    {
        // If the cookie name present in both, equal
        if (!isAttributeName((*it).first))
        {
            if (doesNameExist((*it).first))
                return true;
        }
    }
    return false; 
}

string Cookie::getCookieString(bool bVersionRequired /*=false*/)
{
    // Return a stream based on the Cookie Specification
    string strCookieStream;
    COOKIE_MAP::iterator it;

    if (!isRFC2109Based())
    {
        // Old Cookie Format
        // Form a cookie of the format - 
        //  ie., <name>=<value>[; <name>=<value>] 
        //      [; domain=<domain_name>] [; path=<some_path>]

        // Get the mandatory name-value pairs
        for (it = m_mapNameValue.begin(); it != m_mapNameValue.end(); it++)
        {
            // Ignore the keywords from the map
            if (!isAttributeName((*it).first))
            {
                if (!strCookieStream.empty())
                    strCookieStream += "; ";
                strCookieStream += (*it).first + "=" + (*it).second;
            }
        }
        
        // It's better to send "domain" and "path" info along with the
        // Cookie, but Netscape Spec does not insist on it
        // Get Domain
        // string strVal = getValue("domain");
        //if (strVal.length())
        //    strCookieStream += "; domain=" + strVal;
        // Get Path
        //strVal = getValue("path");
        //if (strVal.length())
        //    strCookieStream += "; path=" + strVal;
    }
    else
    {
        // RFC 2109 based Cookie format
        // Form a cookie of the format - 
        //  ie., $Version="value"; <name>=<"value">[; <name>=<"value">] 
        //      [; $Domain=<"domain_name">] [; $Path=<"some_path">]
        
        // Get Version first 
        string strVal;
        if (bVersionRequired)
        {
            strVal = getValue("version");
            assert(strVal.length());
            strCookieStream += "$Version=\"" + strVal + "\"";
        }

        // Get the mandatory name-value pairs
        for (it = m_mapNameValue.begin(); it != m_mapNameValue.end(); it++)
        {
            // Ignore the keywords from the map
            if (!isAttributeName((*it).first))
            {
                if (bVersionRequired)
                    strCookieStream += "; " ;
                strCookieStream += (*it).first + "=\"" + (*it).second + "\"";
            }
        }
        
        // Get Path
        strVal = getValue("path");
        if (strVal.length())
        {
            strCookieStream += "; $Path=\"" + strVal + "\"";
        }
        // Get Domain
        strVal = getValue("domain");
        if (strVal.length())
        {
            strCookieStream += "; $Domain=\"" + strVal + "\"";
        }
    }
    return strCookieStream;
}

bool Cookie::getAttribValue(const string& strName, string& strVal) const
{ 
    strVal.erase ();
    COOKIE_MAP::const_iterator it = m_mapNameValue.find(strName);
    if (it != m_mapNameValue.end())
    {
       strVal = (*it).second;
       return true;
    }
    else
        return false;
}

void Cookie::getCookieNames(NAME_MAP& mapNames) const
{
    // Returns the non-keyword cookie name-value map.
    // Useful for creating single Cookie objects
    COOKIE_MAP::const_iterator it;
    // Get the mandatory name-value pairs
    for (it = m_mapNameValue.begin(); it != m_mapNameValue.end(); it++)
    {
        if (!isAttributeName((*it).first))
           mapNames.insert(NAME_MAP::value_type((*it).first, (*it).second));
    }
}

void Cookie::updateCookie(const Cookie& objCookie)
{
    // Update the current cookie map with the 
    // passed-in cookie. 
    if (isRFC2109Based() != objCookie.isRFC2109Based())
        return; // Of Different specs

    COOKIE_MAP::const_iterator it;
    for (it = objCookie.m_mapNameValue.begin();
            it != objCookie.m_mapNameValue.end(); it++)
    {
        updateNameValue((*it).first, (*it).second);            
    }
    return; 
}

bool Cookie::isSessionCookie() const
{
    // Session based if
    //      "max-age" does not exist for new proposal
    //      "expires" does not exist as per old cookie specs
    if (isRFC2109Based()) 
        return !doesNameExist("max-age");
    else
        return !doesNameExist("expires");
}

bool Cookie::isCookieValid() const
{ 
    // If no elements in the map, return false
    if (!m_mapNameValue.size())
        return false;
    
    // See if the domain name is as per the specs
    if (!isDomainValid())
        return false;
    
    return true;
}

bool Cookie::hasExpired() const
{
    // Extract the HTTP expiry time (GMT format)
    string strHTTPExpTime;
    if (isRFC2109Based())
    {
        strHTTPExpTime = getValue("max-age");
        if (strHTTPExpTime.empty())
            return false; //OK
        else if (strHTTPExpTime == "0")
            return true;
    }
    else
        strHTTPExpTime = getValue("expires");
    if (!strHTTPExpTime.empty())
    {
        // Convert the Internet time to system time then file time
        SYSTEMTIME objCookieSysTime;
        if (!InternetTimeToSystemTime(strHTTPExpTime.c_str(), &objCookieSysTime, 0))
        {
            assert(false);
            return false;
        }
        FILETIME objCookieFileTime;
        if (!SystemTimeToFileTime( &objCookieSysTime, &objCookieFileTime))
        {
            assert(false);
            return false;
        }

        // Convert local time to to filetime
        SYSTEMTIME  objCurrTime;
        GetSystemTime(&objCurrTime);
        FILETIME objCurrFileTime;
        if (!SystemTimeToFileTime( &objCurrTime, &objCurrFileTime))
        {
            assert(false);
            return false;
        }

        // Check if the cookie has expired
        if (CompareFileTime(&objCookieFileTime, &objCurrFileTime) == -1)
            return true; // expired !
    } // end if()

    // A session cookie
    return false;
}

void Cookie::updateNameValue(const string& strName, const string& strValue)
{
    COOKIE_MAP::iterator it = m_mapNameValue.find (strName);
    if (it !=m_mapNameValue.end())
        (*it).second = strValue;
    else
        insertNameValue(strName, strValue);
}

bool Cookie::doesDomainTailMatch(const string& strReqDomain) const
{
    if (strReqDomain.empty())
        return true;

    string strReqTemp = toLowerStr(strReqDomain); 
    string strDomain = toLowerStr(getDomain());
    if (strDomain == strReqTemp || strDomain.empty())
        return true;

    if (strDomain.length() >= strReqTemp.length())
        return false;

    // Check if of the form strCookieDomain=.foo.com and 
    //  strReqTemp=abc.foo.com
    if (!(strReqTemp[strReqTemp.length()-strDomain.length()] == '.' &&
        strReqTemp.substr(strReqTemp.length()-strDomain.length(),
        strDomain.length()) == strDomain))
            return false;

    // For old Cookie proposal, if strReqTemp ends with strCookieDomain value
    // everything is fine
    if (isRFC2109Based())                   
    {
        // RFC 2109 - A cookie will be rejected, if the request host is a FQDN
        // (Fully Qualified Domain Name - not IP address) and has the form HD,
        // where D is the value of the domain attribute, and H is a string that 
        // contains one or more dots.
        // ie., A Set-Cookie from request-host y.x.foo.com for domain=.foo.com
        // would be rejected, because H is y.x and contains a dot
        string strPart = strReqTemp.substr(0, 
            strReqTemp.length()-strDomain.length());
        if (strPart.find('.', 0) >= 0)
            return true;
        else
            return false;
    }
    return true;
}

bool Cookie::doesPathMatch(const string& strReqPath) const
{
    if (strReqPath.empty())
        return true;

    string strReqTemp = toLowerStr(strReqPath); 
    string strCookiePath = toLowerStr(getPath());
    if (strCookiePath == strReqTemp || strCookiePath=="/"
        || strCookiePath.empty())
        return true;

    if (strCookiePath.length() < strReqTemp.length())
    {
        // Check if of the form strCookiePath=/acme and strReqTemp=/acme/ammo
        if (strReqTemp.substr(0, strCookiePath.length()) == strCookiePath &&
            strReqTemp[strCookiePath.length()] == '/')
            return true;
    }

    return false;
}

void Cookie::constructNameValueMap(const string& strCookie)
{
    // Clear the existing map
    clear();

    // Separate the name=value pairs of the cookie string
    if (strCookie.empty())
        return;

    // Remove the header "set-cookie:", if any
    int nPos = 0;
    string strTempCookie;
    nPos = strCookie.find(':', 0);
    if (nPos >= 0)
    {
        strTempCookie = strCookie.substr(0, nPos+1);
        if (toLowerStr(strTempCookie) != "set-cookie:")
            return; // invalid header
        else
            strTempCookie = strCookie.substr(nPos+1, strCookie.length());
    }
    else // may be, header was removed
        strTempCookie = strCookie;

    bool bFirstName = false;
    string strToken, strName, strVal;
    COOKIE_MAP objTempMap;
    while (!strTempCookie.empty())
    {
        // Get name=value token
        nPos = strTempCookie.find(';', 0);
        if (nPos >= 0)
        {
            strToken = strTempCookie.substr(0, nPos);
            strTempCookie = strTempCookie.substr(nPos+1, strTempCookie.length());
        }
        else
        {
            strToken = strTempCookie;
            strTempCookie.erase();
        }

        // Separate name and value from the token
        nPos = strToken.find('=', 0);
        if (nPos >= 0)
        {
            strName = strToken.substr(0, nPos);
            if (strToken.length() > nPos+1)
                strVal = strToken.substr(nPos+1, strToken.length());
        }
        else
        {
            // No value ? may be a flag like"secure"
            strName = strToken;
        }
        strToken.erase();

        // Cookies are case-insensitive - But let's preserve the case, in case ..
        strName = trimStr(strName);
        strVal  = trimStr(strVal);

        // Update map object
        if (strName.length())
        {
            // If "Version=value" appears as the first token, it's cookie-name
            // not really a version of the cookie spec. Otherwise, update the Cookie
            // Specification flag
            if (!bFirstName)
                bFirstName = true;
            else if (toLowerStr(strName) == "version" && strVal.length())
            {
                if (strVal[0] == '\"')
                    strVal.erase(0, 1);
                if (strVal[strVal.length()-1] == '\"')
                    strVal.erase(strVal.length()-1, 1);
                // If Version 1 or greater, RFC 2109 based Cookie 
                m_nCookieVersion = atoi(strVal.c_str());
            }
            objTempMap.insert(COOKIE_MAP::value_type(strName, strVal));
        }
        
        // Clear the variables for the next iteration
        strName.erase();
        strVal.erase();
    } // end while()

    // Special handling of attribute names and values 
    COOKIE_MAP::iterator it;
    for (it=objTempMap.begin(); it != objTempMap.end(); it++)
    {
        strName = (*it).first;
        strVal  = (*it).second;
        if (isRFC2109Based() && strVal.length())
        {
            // Remove extra """ from values, if any
            if (strVal[0] == '\"')
                strVal.erase(0, 1);
            if (strVal[strVal.length()-1] == '\"')
                strVal.erase(strVal.length()-1, 1);
            // If it's max-age attribute, convert the seconds to Internet 
            // time format for easy expiry checking
            if (toLowerStr(strName) == "max-age")
                strVal = getExpiryInternetTime(atol(strVal.c_str()));
        }
        if (isAttributeName(strName))
            strName = toLowerStr(strName);
 
        // Update the map member
        insertNameValue(strName, strVal);
    }
    objTempMap.clear();
}

string Cookie::getExpiryInternetTime(unsigned long nSeconds) const
{
    // Gets the Current time and add the input seconds to get
    // the expiry time in the Internet Time Format (RFC1123)
    string strInternetTime;

    // If zero seconds, return an expired time
    if (nSeconds <= 0)
        return "Tuesday, 14-June-1988 04:04:55 GMT";

    // Get the localtime and add the expiry seconds
    time_t ltime;
    time( &ltime );
    ltime += nSeconds;

    // Convert local time to to FILETIME
    FILETIME objCurrFileTime;
    LONGLONG llNum = Int32x32To64(ltime, 10000000) + 116444736000000000;
    objCurrFileTime.dwLowDateTime = (DWORD) llNum;
    objCurrFileTime.dwHighDateTime = llNum >> 32;

    // Convert FILETIME to SYSTEMTIME
    SYSTEMTIME objSysTime;
    if (!FileTimeToSystemTime(&objCurrFileTime, &objSysTime))
    {
        assert(false);
        return strInternetTime;
    }
    
    // Convert SYSTEMTIME to Internet Time
    TCHAR* pchTime = new TCHAR [INTERNET_RFC1123_BUFSIZE + 1];
    if (!InternetTimeFromSystemTime(&objSysTime, INTERNET_RFC1123_FORMAT, 
            pchTime, INTERNET_RFC1123_BUFSIZE))
    {
        assert(false);
        delete [] pchTime;
        return strInternetTime;
    }
    pchTime [INTERNET_RFC1123_BUFSIZE] = '\0';

    // Return the Internet Time 
    strInternetTime = pchTime;
    delete [] pchTime;
    return strInternetTime;
}

bool Cookie::isDomainValid() const
{
    // Validate Domain as per the specs
    string strDomain= toLowerStr(getDomain());

    // If no domain specified, OK
    if (strDomain.empty())
        return true;

    // If the domain does not start with a period, problems
    if (strDomain[0] != '.' && isRFC2109Based())
        return false;

    // Get the periods count
    int nPeriodCount = 0;
    for (int i=0; i<strDomain.length(); i++)
    {
        if (strDomain[i] == '.')
            nPeriodCount++;
    }

    // Toplevel domains (7 of them) must have atleast two periods in them
    if (strDomain.length () > 3)
    {
        string strTop = strDomain.substr (strDomain.length()-4, 4);
        if (strTop == ".com" || strTop == ".edu" || strTop == ".net" ||
            strTop == ".org" || strTop == ".gov" || strTop == ".mil" ||
            strTop == ".int")
        {
            // Requires atleast 2 periods in it
            if (nPeriodCount >= 2)
                return true;
            else
                return false;
        }
        // Any other domain requires atleast 3 periods in it
        if (nPeriodCount >= 3)
            return true;
    }

    // Bad Domain
    return false;
}

bool Cookie::isAttributeName(const string& strName) const
{
    // Check id the input word is a reserved word as
    // per the specs
    const char** nameList;
    if (isRFC2109Based())
        nameList = RFC2109_ATTRIBUTES;
    else
        nameList = OLDSPEC_ATTRIBUTES;
    for (int i=0; nameList[i] != NULL; i++)
    {
        if (toLowerStr(strName) == nameList[i])
            return true;
    }
    return false;
}
