///////////////////////////////////////////////////////////////////////
//  CookieManager.cpp: Implementation of CookieManager Class
//  Maintains the Cookies received from the Internet
//
//  Created by Babu George Padamadan
//  Created on Oct 2000
//  Contact : bgeorge@wol.com or babu_george@hotmail.com
//
///////////////////////////////////////////////////////////////////////

#pragma warning (disable:4786)

#include "CookieManager.h"
#include <cassert>

using namespace std;

void CookieManager::deleteCookie(const string& strCookie)
{
	list<Cookie>::iterator it;
	for (it=m_listCookies.begin(); it != m_listCookies.end(); it++)
	{
		if ((*it).isEqual(strCookie))
		{
			m_listCookies.erase(it);
			break;
		}
	}
}

void CookieManager::purgeCookies()
{
	COOKIE_LIST::iterator it;
	for (it=m_listCookies.begin(); it != m_listCookies.end();)
	{
		if (!(*it).isCookieValid() || (*it).hasExpired())
			it = m_listCookies.erase(it);
		else
			it++;
	}
}

void CookieManager::clearSession()
{
	COOKIE_LIST::iterator it;
	for (it=m_listCookies.begin(); it != m_listCookies.end();)
	{
		if ((*it).isSessionCookie() || !(*it).isCookieValid()
            || (*it).hasExpired())
			it = m_listCookies.erase(it);
		else
			it++;
	}
}

string CookieManager::getCookieString(const string& strPath, bool bSecure)
{
	// Create a cookie header to send to the server
	// Format - Cookie: attribute-pair 
	// attribute-pair should be separated by ";" with no ";" at the
	// end of the line. ";" separator is Old and RFC 2109 Compatible

    // If no domain has been specified, error
    assert(!getDomain().empty());
    if (getDomain().empty())
        return "";

	// attribute-pair  is based on the Cookie Specification such as
	// old Cookie or RFC 2109. Cookie class will take care of the 
	// specifications
	bool bAddSemiCRFC = false, bAddSemiCOld = false, bVersionReqd = true;;
	string strRFCCookieStream, strOldCookieStream;

	// sort the list per domain and path
	m_listCookies.sort();

	COOKIE_LIST::iterator it;
	for (it = m_listCookies.begin(); it != m_listCookies.end(); it++)
	{
		// Ignore if it's a not valid cookie or an expired cookie
		if (!(*it).isCookieValid() || (*it).hasExpired())
			continue;

		// Use only those cookies relevant to the current path
		// and domain
		if (!(*it).doesPathMatch(strPath) || !(*it).doesDomainTailMatch(getDomain()))
			continue;

        if ((*it).isRFC2109Based())
        {
		    // If bSecure is true, include secure cookies as well
		    if (bSecure)
		    {	
			    if (bAddSemiCRFC)
                {
				    strRFCCookieStream += "; ";
                    bVersionReqd = false;
                }
			    else
			    {
				    bAddSemiCRFC = true;
				    strRFCCookieStream = "Cookie: ";
			    }

			    // Use all the cookies
			    strRFCCookieStream += (*it).getCookieString(bVersionReqd);
		    }
		    else
		    {
			    // Use only non-secure cookies
			    if (!(*it).isSecure())
			    {
				    if (bAddSemiCRFC)
                    {
					    strRFCCookieStream += "; "; 
                        bVersionReqd = false;
                    }
				    else
				    {
					    bAddSemiCRFC = true;
					    strRFCCookieStream = "Cookie: ";
				    }
				    strRFCCookieStream += (*it).getCookieString(bVersionReqd);
			    }
		    }
        }
        else
        {
            // If bSecure is true, include secure cookies as well
		    if (bSecure)
		    {	
			    if (bAddSemiCOld)
				    strOldCookieStream += "; ";
			    else
			    {
				    bAddSemiCOld = true;
				    strOldCookieStream = "Cookie: ";
			    }

			    // Use all the cookies
			    strOldCookieStream += (*it).getCookieString(bVersionReqd);
		    }
		    else
		    {
			    // Use only non-secure cookies
			    if (!(*it).isSecure())
			    {
				    if (bAddSemiCOld)
					    strOldCookieStream += "; "; 
				    else
				    {
					    bAddSemiCOld = true;
					    strOldCookieStream = "Cookie: ";
				    }
				    strOldCookieStream += (*it).getCookieString(false);
			    }
		    }
        }
	}

	// Internet servers require a "\r\n" at end of the cookie header
	if (bAddSemiCRFC)
		strRFCCookieStream += "\r\n";
    else if (bAddSemiCOld)
        strOldCookieStream += "\r\n";
	return strRFCCookieStream + strOldCookieStream;
}

void CookieManager::updateCookieList(const Cookie& objCookie)
{
	// For efficient Cookie Management, use only one cookie
	// per Cookie object
	NAME_MAP mapNames;
	objCookie.getCookieNames(mapNames);
	if (mapNames.size() == 1)
		addCookie(objCookie);
	else 
	{
		// More than one cookie in the cookie string - let's split them
		// Chances are rare for the following code to be called 
		string strVal;
		NAME_MAP::iterator itMap;
		for (itMap=mapNames.begin(); itMap != mapNames.end(); itMap++)
		{
			Cookie objOneCookie;
			if (objCookie.isRFC2109Based())
			{
				// Insert name
				objOneCookie.insertNameValue ((*itMap).first, (*itMap).second);
				// Insert Version, if any
				strVal = objCookie.getValue ("version");
				if (strVal.length())
					objOneCookie.insertNameValue ("version", strVal);
				// Insert Domain, if any
				strVal = objCookie.getValue ("domain");
				if (strVal.length())
					objOneCookie.insertNameValue ("domain", strVal);
				// Insert Path, if any
				strVal = objCookie.getValue ("path");
				if (strVal.length())
					objOneCookie.insertNameValue ("path", strVal);
				// Insert Max-Age, if any
				strVal = objCookie.getValue ("max-age");
				if (strVal.length())
					objOneCookie.insertNameValue ("max-age", strVal);
				// Insert Comment, if any
				strVal = objCookie.getValue ("comment");
				if (strVal.length())
					objOneCookie.insertNameValue ("comment", strVal);
				// Insert Secure flag, if any
				if (objCookie.doesNameExist ("secure"))
					objOneCookie.insertNameValue ("secure", "");
			}
			else
			{
				// Insert name
				objOneCookie.insertNameValue ((*itMap).first, (*itMap).second);
				// Insert Domain, if any
				strVal = objCookie.getValue ("domain");
				if (strVal.length())
					objOneCookie.insertNameValue ("domain", strVal);
				// Insert Path, if any
				strVal = objCookie.getValue ("path");
				if (strVal.length())
					objOneCookie.insertNameValue ("path", strVal);
				// Insert Expired, if any
				strVal = objCookie.getValue ("expired");
				if (strVal.length())
					objOneCookie.insertNameValue ("expired", strVal);
				// Insert Secure flag, if any
				if (objCookie.doesNameExist ("secure"))
					objOneCookie.insertNameValue ("secure", "");
			}
			addCookie(objOneCookie);
		}
	}
	mapNames.clear();
}

void CookieManager::addCookie(const Cookie& objCookie)
{
	bool bFound = false;
	COOKIE_LIST::iterator it;
	for (it=m_listCookies.begin(); it != m_listCookies.end(); it++)
	{
		// Update the cookie if it exists - isEqual() compares the 
        // cookie names, and domain and path values - not other attributes.
		if ((*it).isEqual(objCookie))
		{
			(*it).updateCookie(objCookie);
			bFound = true;
			break;
		}
	}
	if (!bFound)
		m_listCookies.push_back(objCookie); 
	return;
}