//  tuple_test_bench.cpp  --------------------------------

// Copyright (C) 1999, 2000 Jaakko Jrvi (jaakko.jarvi@cs.utu.fi)
//
// Permission to copy, use, modify, sell and distribute this software is 
// granted provided this copyright notice appears in all copies. 
// This software is provided "as is" without express or implied warranty, 
// and with no claim as to its  suitability for any purpose.

// This program tests tuples in various ways.
// The source contains code, which intentionally produces compile-time errors
// These code sections are guarded by #ifdef directives
// By defining the macros E1 - E11, you can include these sections selectively
// for compilation. If any of these macros is definded, there should be 
// compile time erros.
// ------------------------------------------------------------------


#include "tuple.hpp"

#include <string>
#include <utility>
#include <vector>
#include <algorithm>
#include <functional>
#include <iostream>

#ifndef LL_NO_NAMESPACE
using namespace LL_NAMESPACE;
#endif

using std::make_pair;

template<class T> void dummy(const T& t) {};

class A {};
class B {};
class C {};

  // some random tuple definitions
typedef tuple<int> t1;
typedef tuple<double&, const double&, const double, double*, const double*> t2;
typedef tuple<A, int(*)(char, int), B(A::*)(C&), C> t3;
typedef tuple<std::string, std::pair<A, B> > t4;
typedef tuple<A*, tuple<const A*, const B&, C>, bool, void*> t5;

// An uncopyable class
class no_copy {
  no_copy(const no_copy&) {}
public:
  no_copy() {};
};

no_copy y;

tuple<no_copy&> x = tuple<no_copy&>(y); // ok
#ifdef E1
tuple<no_copy> v1;  // should faild
#endif

char cs[10];
tuple<char(&)[10]> v2(cs);  // ok
#ifdef E2
tuple<char[10]> v3;  // should fail, arrays must be stored as references
#endif


// -tuple construction tests ------------------------------------

// a class without a public default constructor
class no_def_constructor {
  no_def_constructor() {}
public:
  no_def_constructor(std::string) {} // can be constructed with a string
  };


void foo1() {

#ifdef E3
  dummy(tuple<no_def_constructor, no_def_constructor, no_def_constructor>()); // should fail
#endif
  dummy( tuple<no_def_constructor, no_def_constructor, no_def_constructor>(
                      std::string("Jaba"), // ok, since the default 
                      std::string("Daba"),   // constructor is not used
                      std::string("Doo")));
}

void foo2() {
// testing default values
  dummy(tuple<int, double>());
  dummy(tuple<int, double>(1));
  dummy(tuple<int, double>(1,3.14)); 

#ifdef E4
  dummy(tuple<double&>()); // should fail, not defaults for references
  dummy(tuple<const double&>()); // likewise
#endif

  double dd = 5;
  dummy(tuple<double&>(dd)); // ok

#ifdef E5
  dummy(tuple<double&>(dd+3.14)); // should fail, temporary to non-const reference
#endif
  dummy(tuple<const double&>(dd+3.14)); // ok
}



// make_tuple ------------------------------------------


   void foo3() {
    A a; B b;
    const A ca = a;
    make_tuple(cref(a), b);
    make_tuple(ref(a), b);
    make_tuple(ref(a), cref(b));
#ifdef E6
    make_tuple(ref(ca)); // should fail (ca is const, hence ref(ca) is illegal)
#endif
    
    std::cout << make_tuple("Donald", "Daisy"); // should work;
#ifdef E7
    std::make_pair("Doesn't","Work"); // fails
#endif
  }



// - testing element access

void foo4() 
{
  double d = 2.7; 
  A a;
  tuple<int, double&, const A&> t(1, d, a);
  const tuple<int, double&, const A> ct = t;

  int i = get<1>(t);
  int j = get<1>(ct);
  get<1>(t) = 5;
#ifdef E8
  get<1>(ct) = 5; // can't assign to const
#endif

  double e = get<2>(t);

  get<2>(t) = 3.14+i;

#ifdef E9
  get<3>(t) = A(); // can't assign to const
#endif
#ifdef E10
  dummy(get<4>(ct)); // illegal index
#endif

  ++get<1>(t);

}

// testing copy and assignment with implicit conversions between elements
// testing tie

  class AA {};
  class BB : public AA {};
  struct CC { CC() {} CC(const BB& b) {} };
  struct DD { operator CC() const { return CC(); }; };

  void foo5() {
    tuple<char, BB*, BB, DD> t;
    tuple<int, AA*, CC, CC> a(t);
    a = t;
  }

  void foo6() {
    int i; char c; double d;
    tie(i, c, d) = make_tuple(1, 'a', 5.5);
    std::cout << i << " " << c << " " << d << " " << std::endl;
  }

// testing tie
// testing assignment from std::pair
void foo7() {
   int i, j; 
   tie (i, j) = make_pair(1, 2);
   tuple<int, int, float> a;
#ifdef E11
   a = make_pair(1, 2); // should fail, tuple is of length 3, not 2
#endif

   make_tuple(1, 2, 3) = make_tuple(2, 4, 6);
}



// --------------------------------
// ----------------------------
int main() {

  foo1();
  foo2();
  foo3();
  foo4();
  foo5();
  foo6();
  foo7();

  return 0;
}
