// -- type_conversions.hpp -------- ---------------------------------

// Copyright (C) 1999, 2000 Jaakko Jrvi (jaakko.jarvi@cs.utu.fi)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies. 
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice 
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty, 
// and with no claim as to its suitability for any purpose.
//
// For more information, see http://lambda.cs.utu.fi 

#ifndef TYPE_CONVERSIONS_HPP
#define TYPE_CONVERSIONS_HPP

LL_BEGIN_NAMESPACE


// some compile time assertion templates ------------------------

// -- generate error template, referencing to non-existing members of this 
// template is used to produce compilation erros intentionally
template<class T>
class generate_error;

// -- ct_assert<cond>::check leads to a compile time error if cond = false
template <bool Cond> struct ct_assert {
  enum { n = Cond ? 1 : -1 };
  static char check[n];
};

template <bool Cond>
char ct_assert<Cond>::check[n];
// ----------------------------------------------------


// type conversion classes -------------------------------
//   plain_to_c_reference - converts only the plain type to reference to const
//                          identity mapping with all other types
//   plain_to_reference - see above but no const


template<class T>
struct convert {
  typedef const T& plain_to_c_reference;
  typedef T& plain_to_reference;
};

template<class T>
struct convert <const T> {
  typedef const T& plain_to_c_reference;
  typedef const T& plain_to_reference;
};

template<class T>
struct convert <T&> {
  typedef T& plain_to_c_reference;
  typedef T& plain_to_reference;
};

template<class T>
struct convert <const T&> {
  typedef const T& plain_to_c_reference;
  typedef const T& plain_to_reference;
};

template <class T> struct remove_const { typedef T type; };
template <class T> struct remove_const<const T> { typedef T type; };

// -- if construct ------------------------------------------------
// Proposed by Krzysztof Czarnecki and Ulrich Eisenecker

template <bool If, class Then, class Else> struct IF { typedef Then RET; };
template <class Then, class Else> struct IF<false, Then, Else> {
  typedef Else RET;
};
// -----------------------------------------------------------


// reference wrappers --------------------------------------------------

// const_reference_wrapper is used in various functions to  specify 
// that a tuple element should be stored as const reference
template<class T> class  const_reference_wrapper;

template <class T> inline const const_reference_wrapper<T> cref(const T& t);

template<class T>
class  const_reference_wrapper { 
  const T&  x; 
  explicit const_reference_wrapper(const T& t) : x(t) {} 
public:
  operator const T&() const { return x; }
  friend const const_reference_wrapper cref<>(const T& t); 
};

template <class T>
inline const const_reference_wrapper<T> cref(const T& t) { 
  return const_reference_wrapper<T>(t);
}

// reference_wrapper is used to specify that a tuple element should be 
// stored as non-const reference
// A disguise for non-const references. 
// Can only be created via the ref function, which accepts only 
// non-const references. The conversion operator converts the constenss away

template<class T> class reference_wrapper;

template<class T> inline const reference_wrapper<T> ref(T& t);
template<class T> inline const reference_wrapper<T> ref(const T& t);

template<class T>
class reference_wrapper { 
  T& x; 
  explicit reference_wrapper(T& t) : x(t) {} 
public:
  operator T&() const { return x; }

  friend const reference_wrapper ref<>(T& t);
  friend const reference_wrapper ref<>(const T& t);
};

template<class T> 
inline const reference_wrapper<T> ref(T& t) { return reference_wrapper<T>(t); }


template <class T>
inline const reference_wrapper<T> ref(const T& t) { 
  return generate_error<T>::const_reference_not_allowed;  
}


// -- type conversion templates -------------------------------------
// T1 is the default type, specialisations for 
// reference_wrapper, const_reference_wrapper and  array types.
// references to functions require special handling as well

// do not use directly this template directly
// ------------------------------------------------------------------------


// first some helpers
template<class T> struct is_function_reference {
  static const bool test = false;
};

template<class Ret> struct is_function_reference<Ret (...)> {
  static const bool test = true;
};


template<class T1, class T2 = T1> 
struct general_type_conversion {
  typedef typename IF<
    is_function_reference<typename remove_const<T1>::type >::test,
                      T1&,
                      T2>::RET type;
};

template<class T, int n, class Any> struct general_type_conversion<T[n], Any> {
  typedef const T (&type)[n];
};

template<class T, int n, class Any> 
struct general_type_conversion<const T[n], Any> {
  typedef const T (&type)[n];
};

template<class T, class Any> 
struct general_type_conversion<const_reference_wrapper<T>, Any> {
  typedef const T& type;
};
template<class T, class Any> 
struct general_type_conversion<const const_reference_wrapper<T>, Any >{
  typedef const T& type;
};

template<class T, class Any> 
struct general_type_conversion<reference_wrapper<T>, Any >{
  typedef T& type;
};
template<class T, class Any> 
struct general_type_conversion<const reference_wrapper<T>, Any >{
  typedef T& type;
};


// ---------------------------------------------------------------------------
// Use these conversion templates instead of general_type_conversion 

// These must be instantiated with plain or const plain types
// from template<class T> foo(const T& t) : convert_to_xxx<const T>::type
// from template<class T> foo(T& t) : convert_to_xxx<T>::type

// The default is plain type -------------------------
// const T -> const T, 
// T -> T, 
// T[n] -> const T (&) [n]
// references -> compile_time_error
// reference_wrapper<T> -> T&
// const reference_wrapper<T> -> const T&
template<class T>
struct convert_to_plain_by_default {
  typedef typename general_type_conversion<
                     T,
                     T
                   >::type type; 
};
template<class T>
struct convert_to_plain_by_default<T&>; // error

LL_END_NAMESPACE

#endif
