#ifndef BUFFER_INC_
#define BUFFER_INC_

#include <memory>
#include <utility>
#include "TypeTraits.h"
#include "TypeManip.h"

template <class Iter, class T>
inline void try_uninitialized_fill(Iter& beg, Iter end, const T& value)
{
    for (; beg != end; ++beg)
    {
        new(&*beg) T(value);
    }
}

template <class Iter1, class Iter2>
inline void try_uninitialized_copy(Iter1 beg, Iter1 end, Iter2& dest)
{
    for (; beg != end; ++beg, ++dest)
    {
        typedef std::iterator_traits<Iter2>::value_type T;
        new(&*dest) T(*beg);
    }
}

template <class Iter>
void destroy_range(Iter from, Iter to)
{
    for (; from != to; ++from)
    {
        typedef std::iterator_traits<Iter>::value_type T;
        from->~T();
    }
}

template <class Allocator>
bool inplace_reallocate(const Allocator& a, 
    typename Allocator::value_type* p,
    typename Allocator::size_type oldSize,
    typename Allocator::size_type newSize)
{
    return false;
}

template <class Allocator>
void reallocate(const Allocator& a, 
    typename Allocator::value_type*& p,
    typename Allocator::size_type oldSize,
    typename Allocator::size_type newSize)
{
    typedef typename Allocator::value_type T;
    T* pNew = a.allocate(newSize, 0);
    std::uninitialized_copy(p, p + std::min(oldSize, newSize), pNew);
    destroy_range(p, p + oldSize);
    p = pNew;
}

template <class Iter>
void move(Iter first, Iter last, Iter dest)
{
    Iter destEnd = std::copy(first, last, dest);
    size_t n = std::distance(first, last);
    if (dest < first && destEnd > first) first = destEnd;
    if (first < last) destroy_range(first, last);
}

template <class T, class Allocator = std::allocator<T> >
class buffer : private Allocator
{
public:
    // types:
    typedef typename Allocator::reference reference;
    typedef typename Allocator::const_reference const_reference;
    typedef T* iterator; // See 23.1
    typedef const T* const_iterator; // See 23.1
    typedef Allocator::size_type size_type; // See 23.1
    typedef Allocator::difference_type difference_type; // See 23.1
    typedef T value_type;
    typedef Allocator allocator_type;
    typedef typename Allocator::pointer pointer;
    typedef typename Allocator::const_pointer const_pointer;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

    // 23.2.4.1 construct/copy/destroy:
    explicit buffer(const Allocator& a = Allocator())
    : Allocator(a), beg_(0), end_(0)
    {
    }

private:
    enum { copyThrows = TypeTraits::CopyThrows<T>::value != 0 };
    
    // Exceptionless initialization
    void Init(size_type n, const T& value, Loki::Int2Type<false>)
    {
        beg_ = Allocator::allocate(n);
        end_ = beg_ + n;
        iterator i = beg_;
        try_uninitialized_fill(i, end_, value);
    }
    // Initialization in the presence of exceptions
    void Init(size_type n, const T& value, Loki::Int2Type<true>)
    {
        beg_ = Allocator::allocate(n);
        end_ = beg_ + n;
        iterator i = beg_;
        try
        {
            try_uninitialized_fill(i, end_, value);
        }
        catch (...)
        {
            destroy_range(beg_, i);
            Allocator::deallocate(beg_, size());
            throw;
        }
    }
    
public:    
    explicit buffer(size_type n, const T& value = T(),
        const Allocator& a = Allocator())
    : Allocator(a)
    {
        Init(n, value, Loki::Int2Type<copyThrows>());
    }
    
    enum noinit_t { noinit };
    buffer(size_type n, noinit_t, const Allocator& a = Allocator())
    : Allocator(a), beg_(Allocator::allocate(n)), end_(beg_ + n)
    {
    }
    
private:
    // Exceptionless Initialization from range 
    template <class FwdIterator, class Tag>
    void Init(FwdIterator first, FwdIterator last, Tag, 
        Loki::Int2Type<false>)
    {
        const difference_type n = distance(first, last);
        beg_ = Allocator::allocate(n);
        end_ = beg_ + n;
        iterator i = beg_;
        try_uninitialized_copy(first, last, i);
    }
            
    // Initialization from range in the presence of exceptions 
    template <class FwdIterator, class Tag>
    void Init(FwdIterator first, FwdIterator last, Tag, 
        Loki::Int2Type<true>)
    {
        const difference_type n = distance(first, last);
        beg_ = Allocator::allocate(n);
        end_ = beg_ + n;
        iterator i = beg_;
        try
        {
            try_uninitialized_copy(first, last, i);
        }
        catch (...)
        {
            destroy_range(beg_, i);
            Allocator::deallocate(beg_, n);
            throw;
        }
    }
            
    // Initialization from a forward/bidirectional/random range 
    template <class FwdIterator, class Tag>
    void Init(FwdIterator first, FwdIterator last, Tag t)
    {
        Init(first, last, t, Loki::Int2Type<copyThrows>());
    }
            
    // Initialization from an input range 
    template <class FwdIterator>
    void Init(FwdIterator first, FwdIterator last, std::input_iterator_tag)
    {
        beg_ = 0;
        end_ = 0;
        for (; first != last; ++first)
        {
            push_back(*first);
        }
    }

public:
    template <class InputIterator>
    buffer(InputIterator first, InputIterator last, 
        const Allocator& a = Allocator())
    : Allocator(a)
    {
        Init(first, last, 
            typename std::iterator_traits<InputIterator>::iterator_category());
    }

    buffer(const buffer& rhs)
    : Allocator(rhs)
    {
        Init(rhs.begin(), rhs.end(), std::random_iterator_tag);
    }
    
    ~buffer()
    {
        destroy_range(begin(), end());
        Allocator::deallocate(beg_, size());
    }
    
    buffer<T,Allocator>& operator=(const buffer<T,Allocator>& x)
    {
        buffer(x).swap(*this);
    }

private:
    template <class FwdIterator, class Tag>
    void assign_helper(FwdIterator first, FwdIterator last, Tag)
    {
        if (distance(first, last) != size())
        {
            buffer(first, last).swap(*this);
        }
        else
        {
            std::copy(first, last, beg_);
        }
    }
        
    template <class InputIterator>
    void assign_helper(InputIterator first, InputIterator last, 
        std::input_iterator_tag)
    {
        clear();
        for (; first != last; ++first)
        {
            push_back(*first);
        }
    }

public:
    template <class InputIterator>
    void assign(InputIterator first, InputIterator last)
    {
        assign_helper(first, last, 
            typename std::iterator_traits<InputIterator>::iterator_category());
    }
    
    void assign(size_type n, const T& u)
    {
        if (n != size())
        {
            buffer(n, u, *this).swap(*this);
        }
        else
        {
            std::fill(begin(), end(), u);
        }
    }
    
    allocator_type get_allocator() const
    {
        return *this;
    }
    
    // iterators:
    iterator begin()
    {
        return beg_;
    }
    
    const_iterator begin() const
    {
        return beg_;
    }
    
    iterator end()
    {
        return end_;
    }
    
    const_iterator end() const
    {
        return end_;
    }
    
    reverse_iterator rbegin()
    {
        return reverse_iterator(end());
    }
    
    const_reverse_iterator rbegin() const
    {
        return const_reverse_iterator(end());
    }
    
    reverse_iterator rend()
    {
        return reverse_iterator(begin());
    }
    
    const_reverse_iterator rend() const
    {
        return reverse_iterator(begin());
    }
        
    // 23.2.4.2 capacity:
    size_type size() const
    {
        return end_ - beg_;
    }
    
    size_type max_size() const
    {
        return Allocator::max_size();
    }

    void grow_noinit(size_type offset)
    {
        const size_type 
            oldSize = size(), 
            newSize = oldSize + offset;
        Allocator& a = *this;
        reallocate(a, beg_, oldSize, newSize);
        end_ = beg_ + newSize;
    }

private:
    void shrink(size_type offset, bool destroy)
    {
        const size_type 
            oldSize = size(), 
            newSize = (oldSize > offset ? oldSize - offset : 0);

        Allocator& a = *this;
        reallocate(a, beg_, oldSize, newSize, destroy);
        end_ = beg_ + newSize;
    }

public:
    void shrink_nodestroy(size_type offset)
    {
        shrink(offset, false);
    }
    
private:
    // resize for non-throwing types
    void resize_helper(size_type sz, const T& c, Loki::Int2Type<false>)
    {
        const size_type oldSize = size();
        if (sz == oldSize) return;

        if (sz < oldSize) 
        {
            shrink(oldSize - sz, true);
        }
        else
        {
            grow_noinit(sz - oldSize);
            T* pInitLimit = beg_ + oldSize;
            try_uninitialized_fill(pInitLimit, end_, obj);
        }
    }
    
    // resize for throwing types
    void resize_helper(size_type sz, const T& c, Loki::Int2Type<true>)
    {
        const size_type oldSize = size();
        if (sz == oldSize) return;

        if (sz < oldSize) 
        {
            shrink(oldSize - sz, true);
        }
        else
        {
            grow_noinit(sz - oldSize);
            T* pInitLimit = beg_ + oldSize;
            try
            {
                try_uninitialized_fill(pInitLimit, end_, obj);
            }
            catch (...)
            {
                destroy_range(beg_, pInitLimit);
                clear_nodestroy();
                throw;
            }
        }
    }
    
public:
    void resize(size_type sz, const T& c = T())
    {
        resize_helper(sz, c, Loki::Int2Type<copyThrows>());
    }
    
    bool empty() const
    {
        return beg_ == end_; 
    }
    
    // element access:
    reference operator[](size_type n)
    {
        assert(n < size());
        return beg_[n];
    }
    
    const_reference operator[](size_type n) const
    {
        assert(n < size());
        return beg_[n];
    }
    
    const_reference at(size_type n) const
    {
        if (n < size()) throw std::out_of_range();
        return beg_[n];
    }
    
    reference at(size_type n)
    {
        if (n < size()) throw std::out_of_range();
        return beg_[n];
    }
    
    reference front()
    {
        assert(!empty());
        return *beg_;
    }
    
    const_reference front() const
    {
        assert(!empty());
        return *beg_;
    }
    
    reference back()
    {
        assert(!empty());
        return end_[-1];
    }
    
    const_reference back() const
    {
        assert(!empty());
        return end_[-1];
    }

private:
    void push_back(const T& x, Loki::Int2Type<false>)
    {
        buffer temp(size() + 1, noinit);
        T* pInitEnd = temp.beg_;
        try_uninitialized_copy(begin(), end(), pInitEnd);
        new(pInitEnd) T(x);
        temp.swap(*this);
    }
    
    void push_back(const T& x, Loki::Int2Type<true>)
    {
        buffer temp(size() + 1, noinit);
        T* pInitEnd = temp.beg_;
        try
        {
            try_uninitialized_copy(begin(), end(), pInitEnd);
            new(pInitEnd) T(x);
            temp.swap(*this);
        }
        catch (...)
        {
            destroy_range(temp.beg_, pInitEnd);
            temp.clear_nodestroy();
            throw;
        }
    }
    
public:
    // 23.2.4.3 modifiers:
    void push_back(const T& x)
    {
        push_back(x, Loki::Int2Type<copyThrows>());
    }
    
    void pop_back()
    {
        assert(!empty());
        Allocator& a = *this;
        if (!inplace_reallocate(a, beg_, size(), size() - 1))
        {
            buffer(begin(), end() - 1).swap(*this);
        }
    }
    
    iterator insert(iterator position, const T& x)
    {
        size_type n = position - beg_;
        insert(position, 1, x);
        return beg_ + n;
    }

private:
    
    void insert(iterator position, size_type n, const T& x, 
        Loki::Int2Type<false>)
    {
        const size_type n = position - beg_;
        reallocate(beg_, size(), size() + n);
        position = beg_ + n;
        pInitEnd = beg_ + size();
        end_ = pInitEnd + n;
        T* pSwapStart = pInitEnd;
        try_uninitialized_fill(pInitEnd, end_, x);
        swap_ranges(position, position + n, pSwapStart);
    }
    
    void insert(iterator position, size_type n, const T& x,
        Loki::Int2Type<true>)
    {
        T* pInitEnd = 0;
        try
        {
            const size_type n = position - beg_;
            reallocate(beg_, size(), size() + n);
            position = beg_ + n;
            pInitEnd = beg_ + size();
            end_ = pInitEnd + n;
            T* pSwapStart = pInitEnd;
            try_uninitialized_fill(pInitEnd, end_, x);
            swap_ranges(position, position + n, pSwapStart);
        }
        catch (...)
        {
            if (pInitEnd) destroy_range(beg_, pInitEnd);
            beg_ = end_ = 0;
            throw;
        }
    }
    
public:
    void insert(iterator position, size_type n, const T& x)
    {
        insert(position, n, x, Loki::Int2Type<copyThrows>());
    }
    
/*
    template <class InputIterator>
    void insert(iterator position, InputIterator first, InputIterator last)
    {
        size_type n = std::distance(first, last);
        buffer temp(size() + n, noinit);
        iterator pos = std::uninitialized_copy(begin(), position, temp.begin());
        pos = std::uninitialized_copy(first, last, pos);
        std::uninitialized_copy(position, end(), pos);
        temp.swap(*this);
    }
*/
    
    iterator erase(iterator position)
    {
        return erase(position, position + 1);
    }
    
    iterator erase(iterator first, iterator last)
    {
        assert(size() >= last - first);
        
        const size_type n = last - first, leftover = end_ - last;
        move(last, end_, first);
        
        reallocate(beg_, size(), size() - (last - first), false);
    }
    
    void swap(buffer<T,Allocator>& rhs)
    {
        std::swap(beg_, rhs.beg_);
        std::swap(end_, rhs.end_);
    }
    
    void clear()
    {
        const Allocator& myAlloc = *this;
        buffer(myAlloc).swap(*this);
    }
    
    void clear_nodestroy()
    {
        Allocator::deallocate(beg_, size());
        beg_ = end_ = 0;
    }
    
private:
    T* beg_;
    T* end_;
};

#endif